@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 2010, 2016, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, 2018, MariaDB Corporation.
+Copyright (c) 2015, 2019, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -39,7 +39,8 @@ Created 10/13/2010 Jimmy Yang
 	do {								\
 		b[N] = row_merge_read_rec(				\
 			block[N], buf[N], b[N], index,			\
-			fd[N], &foffs[N], &mrec[N], offsets[N]);	\
+			fd[N], &foffs[N], &mrec[N], offsets[N],		\
+			crypt_block[N], space);				\
 		if (UNIV_UNLIKELY(!b[N])) {				\
 			if (mrec[N]) {					\
 				goto exit;				\
@@ -196,6 +197,7 @@ row_fts_psort_info_init(
 	fts_psort_t*		merge_info = NULL;
 	ulint			block_size;
 	ibool			ret = TRUE;
+	bool			encrypted = false;
 
 	block_size = 3 * srv_sort_buf_size;
 
@@ -225,6 +227,10 @@ row_fts_psort_info_init(
 	common_info->merge_event = os_event_create();
 	common_info->opt_doc_id_size = opt_doc_id_size;
 
+	if (log_tmp_is_encrypted()) {
+		encrypted = true;
+	}
+
 	ut_ad(trx->mysql_thd != NULL);
 	const char*	path = thd_innodb_tmpdir(trx->mysql_thd);
 
@@ -261,6 +267,29 @@ row_fts_psort_info_init(
 					ut_align(
 					psort_info[j].block_alloc[i], 1024));
 
+			/* If tablespace is encrypted, allocate additional buffer for
+			encryption/decryption. */
+			if (encrypted) {
+
+				/* Need to align memory for O_DIRECT write */
+				psort_info[j].crypt_alloc[i] =
+					static_cast<row_merge_block_t*>(ut_malloc(
+							block_size + 1024));
+
+				psort_info[j].crypt_block[i] =
+					static_cast<row_merge_block_t*>(
+						ut_align(
+							psort_info[j].crypt_alloc[i], 1024));
+
+				if (!psort_info[j].crypt_block[i]) {
+					ret = FALSE;
+					goto func_exit;
+				}
+			} else {
+				psort_info[j].crypt_alloc[i] = NULL;
+				psort_info[j].crypt_block[i] = NULL;
+			}
+
 			if (!psort_info[j].merge_block[i]) {
 				ret = FALSE;
 				goto func_exit;
@@ -318,6 +347,11 @@ row_fts_psort_info_destroy(
 				if (psort_info[j].block_alloc[i]) {
 					ut_free(psort_info[j].block_alloc[i]);
 				}
+
+				if (psort_info[j].crypt_alloc[i]) {
+					ut_free(psort_info[j].crypt_alloc[i]);
+				}
+
 				mem_free(psort_info[j].merge_file[i]);
 			}
 
@@ -534,9 +568,9 @@ row_merge_fts_doc_tokenize(
 		cur_len += len;
 		dfield_dup(field, buf->heap);
 
-		/* Reserve one byte for the end marker of row_merge_block_t. */
+		/* Reserve one byte for the end marker of row_merge_block_t */
 		if (buf->total_size + data_size[idx] + cur_len
-		    >= srv_sort_buf_size - 1) {
+			>= (srv_sort_buf_size - 1)) {
 
 			buf_full = TRUE;
 			break;
@@ -617,6 +651,7 @@ fts_parallel_tokenization(
 	ibool			processed = FALSE;
 	merge_file_t**		merge_file;
 	row_merge_block_t**	block;
+	row_merge_block_t**	crypt_block;
 	int			tmpfd[FTS_NUM_AUX_INDEX];
 	ulint			mycount[FTS_NUM_AUX_INDEX];
 	ib_uint64_t		total_rec = 0;
@@ -641,13 +676,13 @@ fts_parallel_tokenization(
 	merge_file = psort_info->merge_file;
 	blob_heap = mem_heap_create(512);
 	memset(&doc, 0, sizeof(doc));
-	memset(&t_ctx, 0, sizeof(t_ctx));
 	memset(mycount, 0, FTS_NUM_AUX_INDEX * sizeof(int));
 
 	doc.charset = fts_index_get_charset(
 		psort_info->psort_common->dup->index);
 
 	block = psort_info->merge_block;
+	crypt_block = psort_info->crypt_block;
 	zip_size = dict_table_zip_size(table);
 
 	row_merge_fts_get_next_doc_item(psort_info, &doc_item);
@@ -740,12 +775,14 @@ loop:
 
 		if (!row_merge_write(merge_file[t_ctx.buf_used]->fd,
 				     merge_file[t_ctx.buf_used]->offset++,
-				     block[t_ctx.buf_used])) {
+				     block[t_ctx.buf_used],
+				     crypt_block[t_ctx.buf_used],
+				     table->space)) {
 			error = DB_TEMP_FILE_WRITE_FAILURE;
 			goto func_exit;
 		}
 
-		UNIV_MEM_INVALID(block[t_ctx.buf_used][0], srv_sort_buf_size);
+		UNIV_MEM_INVALID(block[t_ctx.buf_used], srv_sort_buf_size);
 		buf[t_ctx.buf_used] = row_merge_buf_empty(buf[t_ctx.buf_used]);
 		mycount[t_ctx.buf_used] += t_ctx.rows_added[t_ctx.buf_used];
 		t_ctx.rows_added[t_ctx.buf_used] = 0;
@@ -833,13 +870,19 @@ exit:
 			if (merge_file[i]->offset != 0) {
 				if (!row_merge_write(merge_file[i]->fd,
 						merge_file[i]->offset++,
-						block[i])) {
+						block[i],
+						crypt_block[i],
+						table->space)) {
 					error = DB_TEMP_FILE_WRITE_FAILURE;
 					goto func_exit;
 				}
 
-				UNIV_MEM_INVALID(block[i][0],
-						 srv_sort_buf_size);
+				UNIV_MEM_INVALID(block[i], srv_sort_buf_size);
+
+				if (crypt_block[i]) {
+					UNIV_MEM_INVALID(crypt_block[i],
+							 srv_sort_buf_size);
+				}
 			}
 
 			buf[i] = row_merge_buf_empty(buf[i]);
@@ -864,7 +907,10 @@ exit:
 
 		error = row_merge_sort(psort_info->psort_common->trx,
 				       psort_info->psort_common->dup,
-				       merge_file[i], block[i], &tmpfd[i]);
+				       merge_file[i], block[i], &tmpfd[i],
+				       false, 0.0/* pct_progress */, 0.0/* pct_cost */,
+				       crypt_block[i], table->space);
+
 		if (error != DB_SUCCESS) {
 			close(tmpfd[i]);
 			goto func_exit;
@@ -1367,6 +1413,7 @@ row_fts_merge_insert(
 	mrec_buf_t**		buf;
 	int*			fd;
 	byte**			block;
+	byte**			crypt_block;
 	const mrec_t**		mrec;
 	ulint			count = 0;
 	int*			sel_tree;
@@ -1374,6 +1421,7 @@ row_fts_merge_insert(
 	ulint			start;
 	fts_psort_insert_t	ins_ctx;
 	ulint			count_diag = 0;
+	ulint			space;
 
 	ut_ad(index);
 	ut_ad(table);
@@ -1400,6 +1448,8 @@ row_fts_merge_insert(
 	fd = (int*) mem_heap_alloc(heap, sizeof(*fd) * fts_sort_pll_degree);
 	block = (byte**) mem_heap_alloc(
 		heap, sizeof(*block) * fts_sort_pll_degree);
+	crypt_block = (byte**) mem_heap_alloc(
+		heap, sizeof(*block) * fts_sort_pll_degree);
 	mrec = (const mrec_t**) mem_heap_alloc(
 		heap, sizeof(*mrec) * fts_sort_pll_degree);
 	sel_tree = (int*) mem_heap_alloc(
@@ -1420,12 +1470,14 @@ row_fts_merge_insert(
 		offsets[i][0] = num;
 		offsets[i][1] = dict_index_get_n_fields(index);
 		block[i] = psort_info[i].merge_block[id];
+		crypt_block[i] = psort_info[i].crypt_block[id];
 		b[i] = psort_info[i].merge_block[id];
 		fd[i] = psort_info[i].merge_file[id]->fd;
 		foffs[i] = 0;
 
-		buf[i] = static_cast<unsigned char (*)[16384]>(
+		buf[i] = static_cast<mrec_buf_t*>(
 			mem_heap_alloc(heap, sizeof *buf[i]));
+
 		count_diag += (int) psort_info[i].merge_file[id]->n_rec;
 	}
 
@@ -1461,6 +1513,7 @@ row_fts_merge_insert(
 	ins_ctx.fts_table.table_id = table->id;
 	ins_ctx.fts_table.parent = index->table->name;
 	ins_ctx.fts_table.table = index->table;
+	space = table->space;
 
 	for (i = 0; i < fts_sort_pll_degree; i++) {
 		if (psort_info[i].merge_file[id]->n_rec == 0) {
@@ -1473,7 +1526,9 @@ row_fts_merge_insert(
 			if (psort_info[i].merge_file[id]->offset > 0
 			    && (!row_merge_read(
 					fd[i], foffs[i],
-					(row_merge_block_t*) block[i]))) {
+					(row_merge_block_t*) block[i],
+					(row_merge_block_t*) crypt_block[i],
+					space))) {
 				error = DB_CORRUPTION;
 				goto exit;
 			}
